from typing import Optional

import torch
import torch.nn.functional as F
from aqlm.utils import _dequantize_weight, unpack_int_data
from torch import nn


def dequantize_gemm(
    input: torch.Tensor,  #  [..., in_features]
    codes: torch.IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
    codebooks: torch.Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
    scales: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
    bias: Optional[torch.Tensor],
) -> torch.Tensor:
    dequantized_weight = _dequantize_weight(
        unpack_int_data(codes, codebooks.shape[1].bit_length() - 1),
        codebooks,
        scales,
    )
    return F.linear(input, dequantized_weight, bias)



import os
from typing import Optional

import torch
from torch.utils.cpp_extension import load

CUDA_FOLDER = os.path.dirname(os.path.abspath(__file__))
CUDA_KERNEL = load(
    name="codebook_cuda",
    sources=[os.path.join(CUDA_FOLDER, "cuda_kernel.cpp"), os.path.join(CUDA_FOLDER, "cuda_kernel.cu")],
)

torch.library.define(
    "aqlm::code1x16_matmat", "(Tensor input, Tensor codes, Tensor codebooks, Tensor scales, Tensor bias) -> Tensor"
)

torch.library.impl("aqlm::code1x16_matmat", "default", CUDA_KERNEL.code1x16_matmat)


@torch.library.impl_abstract("aqlm::code1x16_matmat")
def code1x16_matmat_meta(input, codes, codebooks, scales, bias):
    return torch.empty(input.shape[:-1] + (codes.shape[0],), device=input.device, dtype=input.dtype)


torch.library.define(
    "aqlm::code2x8_matmat", "(Tensor input, Tensor codes, Tensor codebooks, Tensor scales, Tensor bias) -> Tensor"
)

torch.library.impl("aqlm::code2x8_matmat", "default", CUDA_KERNEL.code2x8_matmat)


@torch.library.impl_abstract("aqlm::code2x8_matmat")
def code2x8_matmat_meta(input, codes, codebooks, scales, bias):
    return torch.empty(input.shape[:-1] + (codes.shape[0],), device=input.device, dtype=input.dtype)



""" Core mathematics for Additive Quantization (AQ): initialization, reconstruction and beam search"""
import math
from typing import Optional

import torch
import torch.nn as nn
from aqlm.inference_kernels import get_backward_pass_kernel, get_forward_pass_kernel
from aqlm.utils import get_int_dtype


class QuantizedLinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        in_group_size: int,
        out_group_size: int,
        num_codebooks: int,
        nbits_per_codebook: int,
        bias=True,
        device=None,
        dtype=None,
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

        assert self.in_features % in_group_size == 0
        assert self.out_features % out_group_size == 0
        num_out_groups = out_features // out_group_size
        num_in_groups = in_features // in_group_size
        self.out_group_size, self.in_group_size = out_group_size, in_group_size
        self.num_codebooks = num_codebooks
        self.nbits_per_codebook = nbits_per_codebook
        self.codebook_size = 2**nbits_per_codebook

        # CODES & CODEBOOKS
        self.codebooks = nn.Parameter(
            torch.empty((num_codebooks, self.codebook_size, out_group_size, in_group_size), **factory_kwargs),
            requires_grad=False,
        )  # [num_codebooks, codebook_size, out_group_size, in_group_size]
        self.codes = nn.Parameter(
            torch.empty(
                (num_out_groups, num_in_groups, num_codebooks),
                device=device,
                dtype=get_int_dtype(nbits_per_codebook),
            ),
            requires_grad=False,
        )  #  [num_out_groups, num_in_groups, num_codebooks]

        # SCALES
        self.scales = nn.Parameter(
            torch.empty((num_out_groups, 1, 1, 1), **factory_kwargs), requires_grad=False
        )  #  [num_out_groups, 1, 1, 1]

        # BIAS
        if bias:
            self.bias = nn.Parameter(torch.empty(out_features, **factory_kwargs), requires_grad=False)
        else:
            self.register_parameter("bias", None)

        # MATMUL_OPS
        self.gemv_op = None
        self.gemm_op = None
        self.use_gemv_rule = None

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.gemv_op is None:
            self.prepare_matmul_op(input)

        if self.use_gemv_rule(input):
            return self.gemv_op.apply(input, self.codes, self.codebooks, self.scales, self.bias)
        else:
            return self.gemm_op.apply(input, self.codes, self.codebooks, self.scales, self.bias)

    def prepare_matmul_op(self, input: torch.Tensor):
        if (
            not input.is_cuda
            and self.codebook_size == 256
            and self.codes.shape[0] == self.out_features // self.out_group_size
        ):
            self.codes.data = torch.permute(self.codes.data, (1, 0, 2)).contiguous()  #  TODO: fix this thing

        self.gemv_op = _get_autograd_matmul_op(
            get_forward_pass_kernel(self.codebooks, False),
            get_backward_pass_kernel(self.codebooks, False),
        )

        self.gemm_op = _get_autograd_matmul_op(
            get_forward_pass_kernel(self.codebooks, True),
            get_backward_pass_kernel(self.codebooks, True),
        )

        self.use_gemv_rule = lambda input: math.prod(input.shape[:-1]) <= 32


def _get_autograd_matmul_op(forward_pass_kernel, backward_pass_kernel):
    class _QuantizedMatmul(torch.autograd.Function):
        @staticmethod
        def forward(
            ctx: torch.Any,
            input: torch.Tensor,
            codes: torch.IntTensor,
            codebooks: torch.Tensor,
            scales: torch.Tensor,
            bias: Optional[torch.Tensor],
        ) -> torch.Tensor:
            ctx.save_for_backward(
                input,
                codes,
                codebooks,
                scales,
                bias,
            )
            return forward_pass_kernel(
                input,
                codes,
                codebooks,
                scales,
                bias,
            )

        @staticmethod
        def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
            input, codes, codebooks, scales, bias = ctx.saved_tensors
            return (
                backward_pass_kernel(
                    grad_output,
                    codes,
                    codebooks,
                    scales,
                    bias,
                ),
                None,
                None,
                None,
                None,
            )

    return _QuantizedMatmul



import warnings
from contextlib import contextmanager
from typing import Callable, Optional

import torch


@contextmanager
def optimize_for_training():
    """
    WARNING: `optimize_for_training` is deprecated. The optimization now happens automatically at runtime.
    OBSOLETE: Use this context manager during model initialization (e.g. `.from_pretrained(...)`) to select inference kernels optimized for larger batch sizes
    """
    warnings.warn("`optimize_for_training` is deprecated. The optimization now happens automatically at runtime.")
    try:
        yield
    finally:
        return


def get_forward_pass_kernel(
    codebooks: torch.Tensor,
    optimize_for_training: bool,
) -> Callable[[torch.Tensor, torch.IntTensor, torch.Tensor, torch.Tensor, Optional[torch.Tensor]], torch.Tensor]:
    num_codebooks, codebook_size, out_group_size, in_group_size = codebooks.shape

    if (optimize_for_training, codebooks.device.type, num_codebooks, codebook_size, out_group_size, in_group_size) == (
        False,
        "cuda",
        1,
        65536,
        1,
        8,
    ):
        from .cuda_kernel import CUDA_FOLDER

        return torch.ops.aqlm.code1x16_matmat
    elif (
        optimize_for_training,
        codebooks.device.type,
        num_codebooks,
        codebook_size,
        out_group_size,
        in_group_size,
    ) == (False, "cuda", 2, 256, 1, 8):
        from .cuda_kernel import CUDA_FOLDER

        return torch.ops.aqlm.code2x8_matmat
    elif (optimize_for_training, codebooks.device.type, out_group_size) == (False, "cuda", 1):
        from .triton_kernel import triton_matmul

        return triton_matmul
    elif (codebooks.device.type, codebook_size, out_group_size) == ("cpu", 256, 1):
        from .numba_kernel import numba_gemm_lut

        return numba_gemm_lut
    else:
        from .dequantization import dequantize_gemm

        return dequantize_gemm


def get_backward_pass_kernel(
    codebooks: torch.Tensor,
    optimize_for_training: bool,
) -> torch.Tensor:
    forward_pass_kernel = get_forward_pass_kernel(
        codebooks=codebooks.transpose(2, 3), optimize_for_training=optimize_for_training
    )

    def _backward_pass_kernel(
        grad_output: torch.Tensor,  #  [..., in_features]
        codes: torch.IntTensor,  #  [num_out_groups, num_in_groups, num_codebooks]
        codebooks: torch.Tensor,  #  [num_codebooks, codebook_size, out_group_size, in_group_size]
        scales: torch.Tensor,  #  [num_out_groups, 1, 1, 1]
        bias: Optional[torch.Tensor],
    ) -> torch.Tensor:
        return forward_pass_kernel(
            grad_output.contiguous(),
            codes.transpose(0, 1).contiguous(),
            codebooks.transpose(2, 3).contiguous(),
            scales.transpose(0, 1).transpose(2, 3).contiguous(),
            None,
        )

    return _backward_pass_kernel
